#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "label.h"
#include "parse.h"
#include "linebuf.h"

#define EJ(e,j) j

int generate_pc88_d88(char *, unsigned char *, int, int, int);
int generate_mz2000_d88(char *, unsigned char *, int, int, int, char *);
int parse_file(char *);

FILE *pass2 = NULL;
char *labeldef_by_cmdline = NULL;

static int disk_type;
#define DISK_TYPE_2D  0
#define DISK_TYPE_2DD 1

int get1line(char *buf, parseinfo *pi) {
	if (pi->srcfile) {
	    /* read 1 line from file */
	    if (fgets(buf, 1024, pi->srcfile)) return 1;
	} else if (pi->srclinebuf) {
	    /* read 1 line from line buffer */
	    char *macroext, *p, *p1, *p2, *p3;
	    int siz = 1023, len, err = 0;
	    unsigned long la;
	    macroext = get_linebuf(&pi->srclinebuf);
	    if (macroext) {
		for (p1 = macroext, p2 = buf; (p = strchr(p1, MACRO_ARG_CHAR)) != NULL;) {
		    if (p[1] < '1' || p[1] > '9') { /* '%' as operator */
			len = p-p1+1;
			if (len > siz) { err++; break; }
			strncpy(p2, p1, len);
			p1 += len;
			p2 += len;
			siz -= len;
			continue;
		    }
		    len = p-p1;
		    if (len > siz) { err++; break; }
		    strncpy(p2, p1, len); /* before '%x' */
		    p1 += len + 2;
		    p2 += len;
		    siz -= len;
		    la = pi->macroargs[(int)(p[1] - '1')];
		    if (la) {
			p3 = get_linebuf(&la);
			len = strlen(p3);
			if (len > siz) { err++; break; }
			strncpy(p2, p3, len);
			p2 += len;
			siz -= len;
		    }
		}
		if (!err) {
		    len = strlen(p1);
		    if (len > siz) {
			err++;
		    } else {
			strcpy(p2, p1);
			return 1;
		    }
		}
		printf("Error %s     : s܂\n", get_string(pi->filename));
	    }
	}
	return 0;
}

int parse_text(parseinfo *pi) {
	int error_found;
	int r;
	char buf[1024];
	error_found = 0;
	while (get1line(buf, pi)) {
	    r = parse(buf, pi);
	    if (r == PRES_OK)
		pi->linenum++;
	    else if (r == PRES_END) break;
	    else if (r == PRES_ERROR) {
		error_found++;
		pi->linenum++;
	    } else if (r == PRES_INCLUDE) {
		if (parse_file(pi->buf)) r = PRES_OK;
		else {
		    r = PRES_ERROR;
		    break;
		}
		pi->linenum++;
	    } else if (r == PRES_BINCLUDE) {
		FILE *fpi;
		int c;
		if ((fpi = fopen(pi->buf, "rb")) != NULL) {
		    while ((c = fgetc(fpi)) != EOF) {
			if (!putbin(c)) {
			    printf("Error %s%5d: File %s is too large\n", get_string(pi->filename), pi->linenum, pi->buf);
			    break;
			}
		    }
		    fclose(fpi);
		    if (c == EOF) r = PRES_OK; else break;
		} else {
		    printf("Error %s%5d: File %s is not found\n", get_string(pi->filename), pi->linenum, pi->buf);
		    break;
		}
		pi->linenum++;
	    } else if (r == PRES_MACRO) {
		label *l;
		char macro_name[256];
		int rr, erl;
		rr = 0;
		erl = pi->linenum;
		strcpy(macro_name, pi->buf);
		l = getlabelptr(getlabel(macro_name));
		if (pi->pass2) {
		    /* skip inside macro-endm */
		    pi->linenum++;
		    while (get1line(buf, pi)) {
			pi->linenum++;
			if (parse_endm(buf, pi)) break;
		    }
		    if (pi->rept_cnt > 0) goto proc_rept;
		    continue;
		}
		pi->linenum++;
		if (get1line(buf, pi) && (rr = parse_endm(buf, pi)) == 0) {
		    l->value = (unsigned short) add_linebuf(buf);
		    if (!l->value) break;
		    pi->linenum++;
		}
		if (rr < 0) { error_found++; break; }
		while (get1line(buf, pi)) {
		    add_linebuf(buf);
		    rr = parse_endm(buf, pi);
		    pi->linenum++;
		    if (rr == 1) break;
		    if (rr < 0) {
			error_found++;
			pi->rept_cnt = 0;
			break;
		    }
		}
		end_addline();
		if (rr < 0) { error_found++; break; }
		if (rr == 0) {
		    printf("Error %s%5d: ΉENDM݂܂\n", get_string(pi->filename), erl);
		    error_found++;
		    break;
		}
		r = PRES_OK;
proc_rept:
		if (pi->rept_cnt > 0) {
		    /* REPT-ENDM */
		    parseinfo pinfo2;
		    pinfo2 = *pi;
		    pinfo2.linenum  = 1;
		    pinfo2.srcfile  = NULL;
		    pinfo2.ifstt    = IFSTT_NONE;
		    pinfo2.ifsp     = 0;
		    pinfo2.rept_cnt = 0;
		    for (; pi->rept_cnt; pi->rept_cnt--) {
			pinfo2.srclinebuf = l->value;
			r = parse_text(&pinfo2);
			if (r != PRES_OK) {
			    error_found++;
			    break;
			}
		    }
		}
	    }
	}
	if (error_found) r = PRES_ERROR;
	return r;
}

int parse_file(char *fn) {
	FILE *fp;
	parseinfo pinfo;
	int r;
	int error_found = 0;
	static int nestlevel = 0;

	fp = fopen(fn, "r");
	if (!fp) {
	    printf("File \'%s\' cannot open\n", fn);
	    return 0;
	}
	nestlevel++;
	pinfo.linenum = 1;
	pinfo.filename = add_string(fn);
	pinfo.srcfile = fp;
	pinfo.ifstt = IFSTT_NONE;
	pinfo.ifsp = 0;
	pinfo.pass2 = pass2;
	pinfo.rept_cnt = 0;
	pinfo.srclinebuf = 0;

	r = parse_text(&pinfo);

	fclose(fp);
	if (r == PRES_OK) {
	    if (nestlevel == 1)
		printf("Warning: %s: EOF is found before 'END' command\n", fn);
	    r = PRES_END;
	}
	nestlevel--;
	return (r == PRES_END);
}

void quit(int r) {
	tini_labeltable();
	tini_linebuf();
	exit(r);
}

void usage(void) {
	//      01234567890123456789012345678901234567890123456789012345678901234567890123456789
	printf("Z80AS -- Z80 Absolute Assembler ver 0.12 by youkan\n\n");
	printf("Usage: Z80AS [options] sourcefile [,sourcefile...]\n");
	printf("  -oxxx Set output file name\n");
	printf("  -i    Output with Intel HEX format\n");
	printf("  -mxxx Output MZT file (xxx is filename in MZT image)\n");
	printf("  -2xxx Output D88 file for MZ-2000 (xxx is boot disk name)\n");
	printf("  -8    Output D88 file for PC-8801\n");
	printf("  -d    Just dump the result\n");
	printf("  -l    Output label definition list (.lst)\n");
	printf("  -x    Output assemble list (.als)\n");
	printf("  -D    Define a label (ex. -Dlabel=1234h)\n");
	printf("  -T    Truncate MZT file if it exceeds D000h\n");
	printf("  -Y    Output symbol table file (.sym)\n");
	quit(0);
}

int parse_files(unsigned long infiles) {
	char *fn;

	if (labeldef_by_cmdline != NULL) {
	    parseinfo pinfo;
	    pinfo.linenum = 1;
	    pinfo.filename = add_string("in -D option");
	    pinfo.ifstt = IFSTT_NONE;
	    pinfo.ifsp = 0;
	    pinfo.pass2 = pass2;
	    pinfo.rept_cnt = 0;
	    if (parse(labeldef_by_cmdline, &pinfo) != PRES_OK) return 0;
	}

	while ((fn = get_linebuf(&infiles)) != 0) {
	    if (!parse_file(fn)) return 0;
	}
	return 1;
}

#define FNLEN 256
int copyfn(char *fnbuf, char *fn) {
	strncpy(fnbuf, fn, FNLEN);
	return !fnbuf[FNLEN-1];
}
int copyext(char *fnbuf, char *ext) {
	char c, *p, *lastp;
	for (p = fnbuf, lastp = 0; (c = *p) != 0; p++) {
	    if (c == '.') lastp = p;
	}
	if (!lastp) lastp = p;
	if (lastp - fnbuf > FNLEN - 5) return 0;
	strcpy(lastp, ext);
	return 1;
}

#define RAW_MODE 0
#define MZT_MODE 1
#define INTEL_HEX_MODE 2
#define DUMP_MODE 3
#define PC88_MODE 4
#define MZ2000_MODE 5
#define MZ2500_MODE 6
void main(int argc, char *argv[]) {
	long li;
	label *lp;
	char *infile  = 0;
	char *outfile = 0;
	FILE *fp;
	char *ext = ".bin";
	char buf[FNLEN];
	char mzth[128];
	int i, j;
	int outputmode = 0;
	int outlist = 0;
	int outlist_local = 0;
	int outlist_global = 0;
	int outasm = 0;
	int outsym = 0;
	int truncateD000 = 0;
	int sizD000 = 0;
	int siz, loadaddr, execaddr;
	unsigned char *bin;
	unsigned long infiles;
	int retcode = 0;
	char *labelfilter;
	int labelfilterlen = 0;

	if (argc == 1) usage();

	if (!init_labeltable()) {
	    printf("Can't initialize label table\n");
	    quit(-1);
	}
	if (!init_linebuf()) {
	    printf("Can't initialize macro buffer\n");
	    quit(-1);
	}

	for (i=1; i<argc; i++) {
	    char *p;
	    p = argv[i];
	    if (*p == '-') {
		/* options */
		switch (p[1]) {
		    case 'o':
			if (p[2]) outfile = &p[2];
			else if (i < argc) outfile = argv[++i];
			else usage();
			if (!copyfn(buf, outfile)) {
			    printf("Output filename too long\n");
			    quit(-1);
			}
			outfile = buf;
			break;
		    case 'm': {
			int islow;
			if (p[2]) p = &p[2];
			else if (i < argc) p = argv[++i];
			else usage();
			ext = ".mzt";
			mzth[0] = 0x01; /* binary */
//			mzth[1] = 0x20; /* if saved from 1Z-009A, 1st char of filename is ' ' */
			islow = 0;
			for (j=1/*2*/; j<0x12; j++) {
			    char c;
			    c = *p++;
			    if (c >= 'a' && c <= 'z') {
				c -= 0x20;
				if (!islow) {
				    if (j<0x11) {
					mzth[j++] = 0x05;
				    } else c=0;
				    islow = 1;
				}
			    } else if (islow) {
				if (j<0x11) mzth[j++] = 0x05; else c=0;
				islow = 0;
			    }
			    if (c == 0) {
				mzth[j++] = 0x0D;
				for (; j<0x12; j++) mzth[j] = 0x20;
			    } else
				mzth[j] = c;
			}
			outputmode = MZT_MODE;
			break;
		    }
		    case 'i':
			outputmode = INTEL_HEX_MODE;
			ext = ".hex";
			break;
		    case '8':
			outputmode = PC88_MODE;
			ext = ".d88";
			break;
		    case '2': {
			int m = 0;
			if (p[2] == '5') { p++; m = 1; }
			if (p[2]) p = &p[2];
			else if (i < argc) p = argv[++i];
			else usage();
			outputmode = m ? MZ2500_MODE : MZ2000_MODE;
			ext = ".d88";
			strncpy(mzth, p, 10); /* keep boot file name */
			mzth[10] = 0;
			break;
		    }
		    case 'd':
			outputmode = DUMP_MODE;
			break;
		    case 'l':
			outlist = 1;
			if (p[2] == 'l') outlist_local = 1;
			else if (p[2] == ':') {
			    labelfilter = &p[3];
			    labelfilterlen = strlen(labelfilter);
			} else if (p[2] == 'g') outlist_global = 1;
			break;
		    case 'x':
			outasm = 1;
			break;
		    case 'T':
			truncateD000 = 1;
			break;
		    case 'D':
			if (!p[2]) {
			    printf("Error in -D option\n");
			    quit(-1);
			}
			labeldef_by_cmdline = &p[2];
			break;
		    case 'Y':
			outsym = 1;
			break;
		    default:
			printf("Illegal option: -%c\n", p[1]);
			quit(-1);
		}
	    } else {
		/* source file name */
		if (!infile) {
		    infile = p;
		    infiles = add_linebuf(p);
		} else add_linebuf(p);
	    }
	}
	if (!infile) usage();
	end_addline();
	if (!outfile) {
	    if (!copyfn(buf, infile) || !copyext(buf, ext)) {
		printf("Input filename is too long\n");
		quit(-1);
	    }
	    outfile = buf;
	}

	start_asm();
	if (parse_files(infiles)) {

	    char *openerrmes = "Cannot open output file\n";

	    if (!finish_asm()) {
		retcode = -1;
		goto xit;
	    }

	    bin = getbininfo(&siz, &loadaddr, &execaddr);
	    if (!siz) {
		printf("Binary is empty. No output file is generated.\n");
		retcode = -1;
		goto xit;
	    }

	    switch (outputmode) {

		case RAW_MODE:
		    fp = fopen(outfile, "wb");
		    if (!fp) {
			printf(openerrmes);
			retcode = -1;
			goto xit;
		    }
		    fwrite(bin, 1, siz, fp);
		    fclose(fp);
		    break;

		case MZT_MODE:
		    if (truncateD000) {
			if (loadaddr + siz > 0xD000) {
			    sizD000 = loadaddr + siz - 0xD000;
			    siz -= sizD000;
			}
		    }
		    mzth[0x12] = (unsigned char)( siz            & 0x00ff);
		    mzth[0x13] = (unsigned char)((siz      >> 8) & 0x00ff);
		    mzth[0x14] = (unsigned char)( loadaddr       & 0x00ff);
		    mzth[0x15] = (unsigned char)((loadaddr >> 8) & 0x00ff);
		    mzth[0x16] = (unsigned char)( execaddr       & 0x00ff);
		    mzth[0x17] = (unsigned char)((execaddr >> 8) & 0x00ff);
		    for (i=0x18; i<128; i++) mzth[i] = 0;
		    fp = fopen(outfile, "wb");
		    if (!fp) {
			printf(openerrmes);
			retcode = -1;
			goto xit;
		    }
		    fwrite(mzth, 1, 128, fp);
		    fwrite(bin, 1, siz, fp);
		    fclose(fp);
		    if (sizD000) {
			fp = fopen("D000.bin", "wb");
			fwrite(bin+siz, 1, sizD000, fp);
			fclose(fp);
		    }
		    break;

		case INTEL_HEX_MODE: {
		    int i,j;
		    unsigned char c, chksum;
		    i = 0;
		    fp = fopen(outfile, "w");
		    if (!fp) {
			printf(openerrmes);
			retcode = -1;
			goto xit;
		    }
		    while (siz) {
			chksum = 0;
			j = (siz < 16) ? siz : 16;
			siz -= j;
			chksum -= (unsigned char)j;
			chksum -= (unsigned char)(((loadaddr+i) >> 8) & 0x0000ff);
			chksum -= (unsigned char)( (loadaddr+i)       & 0x0000ff);
			fprintf(fp, ":%02X%04X00", j, loadaddr+i);
			i += j;
			for (; j; j--) {
			    c = (*bin++ & 0x0000ff);
			    chksum -= c;
			    fprintf(fp, "%02X", c);
			}
			fprintf(fp, "%02X\n", chksum);
		    }
		    fprintf(fp, ":00000001FF\n");
		    fclose(fp);
		}

		case PC88_MODE:
		    disk_type = DISK_TYPE_2D;
		    if (!generate_pc88_d88(outfile, bin, siz, loadaddr, execaddr)) {
			retcode = -1;
			goto xit;
		    }
		    break;

		case MZ2000_MODE:
		    disk_type = DISK_TYPE_2D;
		    if (loadaddr != 0 || execaddr != 0) {
			printf("Warning: [hAhXEsAhX0000hł͂܂\n");
		    }
		    if (!generate_mz2000_d88(outfile, bin, siz, loadaddr, execaddr, mzth)) {
			retcode = -1;
			goto xit;
		    }
		    break;

		case MZ2500_MODE:
		    disk_type = DISK_TYPE_2DD;
		    if ((loadaddr & 0x1FFF) != 0) {
			printf("Warning: [hAhX8KoCgEɑĂ܂\n");
		    }
		    if ((loadaddr >= 0xE000) || (loadaddr+siz >= 0xE000)) {
			printf("Warning: oCi0000h`DFFFh͈̔͂𒴂Ă܂\n");
		    }
		    if (!generate_mz2000_d88(outfile, bin, siz, loadaddr, execaddr, mzth)) {
			retcode = -1;
			goto xit;
		    }
		    break;

		case DUMP_MODE: {
		    /*dbg: dump result*/
		    int i,j;
		    j = loadaddr & 7;
		    if (j) {
			printf("%04X:", loadaddr & ~7);
			for (i=0; i<j; i++) printf("   ");
		    }
		    for (i=0; i<siz; i++) {
			if (j==0) printf("%04X:",loadaddr+i);
			printf(" %02X", *bin++ & 0x0000ff);
			if (++j >= 8) {j=0; printf("\n");}
		    }
		    if (j) printf("\n");
		}

	    }

	} else {
	    quit_asm();
	    retcode = -1;
	    goto xit;
	}

	if (outlist) {
	    labellist *llp = NULL, *llp0, *lltmp, **llpn;
	    char *p;
	    /* output labellist */
	    copyext(outfile, ".lst");
	    fp = fopen(outfile, "wb");
	    if (fp) {
		/* sort labels */
		topmost_namespace();
		do {
		    for (li=1; (lp = getlabelptr(li)) != NULL; li++) {
			if (lp->defstate != LABEL_DEFINED) continue;
			if (lp->reftype & LF_HIDE) continue;
			if (outlist_global && !(lp->reftype & LF_GLOBAL)) continue;
			p = getlabelstr(li);
			if (p[0]=='@') continue; /* local label */
			if (labelfilterlen) {
			    if (strncmp(p, labelfilter, labelfilterlen)) continue;
			}
			lltmp = malloc(sizeof(labellist));
			lltmp->li = li;
			lltmp->next = NULL;
			for (llpn = &llp, llp0 = llp; llp0; llpn = &(llp0->next), llp0 = llp0->next) {
			    if (strcmp(p, getlabelstr(llp0->li)) < 0) {
				lltmp->next = llp0;
				*llpn = lltmp;
				break;
			    }
			}
			if (llp0 == NULL) *llpn = lltmp;
		    }
		    for (llp0 = llp; llp0; llp0 = llp0->next) {
			p = getlabelstr_with_namespace(llp0->li);
			j = 2-(strlen(p)/8);
			fprintf(fp, "%s", p);
			if (j <= 0) fprintf(fp, " ");
			else for (; j > 0; j--) fprintf(fp, "\t");
			fprintf(fp, "EQU\t0%04XH\n", (getlabelptr(llp0->li))->value);
		    }
		    /* dispose labellist */
		    while (llp) {
			lltmp = llp;
			llp = lltmp->next;
			free(lltmp);
		    }
		} while (outlist_local && next_namespace());
		fclose(fp);
	    }
	}

	if (outsym) {
	    int cnt;
	    char *p;
	    /* output symbol table */
	    copyext(outfile, ".sym");
	    fp = fopen(outfile, "wb");
	    if (fp) {
		cnt = 0;
		topmost_namespace();
//		do {
		    for (li=1; (lp = getlabelptr(li)) != NULL; li++) {
			if (lp->defstate != LABEL_DEFINED) continue;
			if (lp->reftype & (LF_HIDE|LF_REDEFINABLE)) continue;
			p = getlabelstr(li);
			if (p[0]=='@') continue; /* local label */
			p = getlabelstr_with_namespace(li);
			fprintf(fp, "%04X %s\t", lp->value, p);
			cnt = (cnt+1) & 3;
			if (cnt == 0) fprintf(fp, "\r\n");
		    }
//		} while (next_namespace());
		fclose(fp);
	    }
	}

	if (outasm) {
	    copyext(outfile, ".als");
	    pass2 = fopen(outfile, "wb");
	    start_asm();
	    parse_files(infiles);
	    quit_asm();
	    fclose(pass2);
	}

xit:
	tini_labeltable();
	tini_linebuf();
	exit(retcode);
}

/*
    PC-8801p ȈD88
 */
unsigned char iplbin[] = {
	0x18, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xF3, 0x21, 0x16, 0xC0, 0x11, 0x1E, 0xF2, 0x01,
	0x82, 0x00, 0xD5, 0xED, 0xB0, 0xC9, 0x31, 0xC0,
	0xF2, 0x3E, 0x39, 0xD3, 0x31, 0x3E, 0xA8, 0xD3,
	0x32, 0x2A, 0x02, 0xC0, 0x7D, 0x32, 0x7B, 0xF2,
	0x7C, 0xD3, 0x70, 0xED, 0x5B, 0x04, 0xC0, 0x1B,
	0x1C, 0x14, 0x0E, 0x01, 0x3E, 0x02, 0xCD, 0xCC,
	0x37, 0x1E, 0x10, 0x7A, 0xBB, 0x30, 0x01, 0x5F,
	0x7B, 0xCD, 0xD2, 0x37, 0xAF, 0xCD, 0xD2, 0x37,
	0x79, 0xCD, 0xD2, 0x37, 0x3E, 0x01, 0xCD, 0xD2,
	0x37, 0x3E, 0x12, 0xCD, 0xCC, 0x37, 0xC5, 0xCD,
	0x77, 0xF2, 0x1D, 0x20, 0xFA, 0xC1, 0x0C, 0x7A,
	0xD6, 0x10, 0x57, 0x28, 0x02, 0x30, 0xCD, 0x3E,
	0x3B, 0xD3, 0x31, 0x2A, 0x06, 0xC0, 0xE9, 0x01,
	0xFC, 0x00, 0x21, 0x00, 0x80, 0x3E, 0x0B, 0xD3,
	0xFF, 0xDB, 0xFE, 0x0F, 0x30, 0xFB, 0x3E, 0x0A,
	0xD3, 0xFF, 0xED, 0xA2, 0x3E, 0x0D, 0xD3, 0xFF,
	0xDB, 0xFE, 0x0F, 0x38, 0xFB, 0xED, 0xA2, 0x3E,
	0x0C, 0xD3, 0xFF, 0x20, 0xE0, 0xD3, 0x78, 0xC9,
};

#define HEADERSIZE (17+9+1+1+4+(4*164))
#define TTBLENTRY  (17+9+1+1+4)

#define DISK_SIZE_2D     (0x20 + (4*164) + ((16 + 256) * 16) * 80)
#define DISK_SIZE_2DD    (0x20 + (4*164) + ((16 + 256) * 16) * 160)

int create_new_d88(char *fn) {
	unsigned char hbuf[0x20];
	unsigned long tbuf[164];
	unsigned char sbuf[16+256];
	FILE *fp;
	int i, j, k, r;
	int disk_size, track_num;
	long siz;

	r = 0;
	fp = fopen(fn, "wb");
	if (fp == NULL) goto xit;

	switch (disk_type) {
	    case DISK_TYPE_2D : disk_size = DISK_SIZE_2D ; track_num =  80; break;
	    case DISK_TYPE_2DD: disk_size = DISK_SIZE_2DD; track_num = 160; break;
	    default: fprintf(stderr, "m̃fBXN^Cv\n"); goto xit;
	}

	/* wb_쐬 */
	for (i=0; i<0x1C; i++) hbuf[i] = 0;
	hbuf[i++] = (unsigned char)( disk_size        & 0x000000FF);
	hbuf[i++] = (unsigned char)((disk_size >>  8) & 0x000000FF);
	hbuf[i++] = (unsigned char)((disk_size >> 16) & 0x000000FF);
	hbuf[i  ] = (unsigned char)((disk_size >> 24) & 0x000000FF);
	siz = fwrite(&hbuf[0], 1, 0x20, fp);
	if (siz != 0x20) goto xit;

	/* gbNItZbge[u쐬 */
	for (i=0; i<track_num; i++) tbuf[i] = 0x2B0 + ((16 + 256) * 16) * i;
	for (   ; i<164;       i++) tbuf[i] = 0;
	siz = fwrite(&tbuf[0], 4, 164, fp);
	if (siz != 164) goto xit;

	/* ZN^f[^쐬 */
	for (i=0; i<16+256; i++) sbuf[i] = 0;
	sbuf[3] =  1;	// ZN^TCY = 256
	sbuf[4] = 16;	// gbÑZN^
	sbuf[15] = 1;	// ZN^TCY = 256
	for (i=0; i<40; i++) {
	    sbuf[0] = i;	// ID C
	    for (j=0; j<2; j++) {
		sbuf[1] = j;	// ID H
		for (k=1; k<=16; k++) {
		    sbuf[2] = k;	// ID R
		    siz = fwrite(&sbuf[0], 1, 16+256, fp);
		    if (siz != 16+256) goto xit;
		}
	    }
	}
	r = 1;
xit:
	fclose(fp);
	return r;
}

int write_d88(char *d88fn,
	      int start_track, int start_sector,
	      unsigned char *binbuf, int binsiz) {
	unsigned char *binp, *d88buf;
	char *errmsg = NULL;
	long d88siz, rsiz;
	int t, s;
	FILE *fp;
//fprintf(stderr, "%s: Track %d, Sector %d\n", d88fn, start_track, start_sector);
	/* D88t@Cǂݍ */
	fp = fopen(d88fn, "rb");
	if (fp == NULL) {
	    if (!create_new_d88(d88fn)) {
		errmsg = "D88t@C쐬ł܂łBf܂";
		goto xit;
	    }
	    fp = fopen(d88fn, "rb");
	}
	if (fp == NULL) {
		free(binbuf);
		fprintf(stderr, "'%s'ǂݍ߂܂\n", d88fn);
		return 0;
	}
	fseek(fp, 0, SEEK_END);
	d88siz = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	d88buf = malloc(d88siz);
	if (d88buf == NULL) {
		free(binbuf);
		fclose(fp);
		fprintf(stderr, "mۂł܂\n");
		return 0;
	}
	rsiz = fread(d88buf, 1, d88siz, fp);
	fclose(fp);
	if (rsiz != d88siz) {
		free(binbuf);
		free(d88buf);
		fprintf(stderr, "'%s'̓ǂݍݒɃG[܂\n", d88fn);
		return 0;
	}

	binp = binbuf;
	for (t=start_track; t<164; t++) {
		unsigned char *sbuf;
		long tpos, ssiz, snum, siz;
		tpos = (d88buf[TTBLENTRY+t*4]      ) |
		       (d88buf[TTBLENTRY+t*4+1]<< 8) |
		       (d88buf[TTBLENTRY+t*4+2]<<16) |
		       (d88buf[TTBLENTRY+t*4+3]<<24);
		if (!tpos) continue;
		sbuf = &d88buf[tpos];
//fprintf(stdout, "Track %d\n", t);
		/* ZN^ǂ */
		snum = sbuf[0x04] | (sbuf[0x05]<<8);
		if (t != (sbuf[0]*2+sbuf[1])) {
		    errmsg = "ʏȊÕtH[}bgɂ͑ΉĂ܂";
		    goto xit;
		}
		for (s=start_sector; s<=snum; s++) {
//fprintf(stderr, "tpos = %08X\n", tpos);
//fprintf(stdout, "Sector %d\n", s);
		    sbuf = &d88buf[tpos];
		    while (s != sbuf[2]) {
			sbuf += 16 + ssiz;
		    }
//fprintf(stderr, "C:%d H:%d R:%d N:%d\n", sbuf[0], sbuf[1], sbuf[2], sbuf[3]);
		    if (t != (sbuf[0]*2+sbuf[1]) || s != sbuf[2]) {
			errmsg = "ʏȊÕtH[}bgɂ͑ΉĂ܂";
			goto xit;
		    }
		    ssiz = (sbuf[0x0e] | (sbuf[0x0f]<<8));
		    siz = (ssiz < binsiz) ? ssiz : binsiz;
//fprintf(stderr, "siz = %d\n", siz);
		    memcpy(&sbuf[16], binp, siz);
		    binsiz -= siz;
		    binp += siz;
		    if (binsiz == 0) break;
		    tpos += 16 + siz;
//fprintf(stderr, "tpos = %08X\n", tpos);
		}
		start_sector = 1;
		if (binsiz == 0) break;
	}
	if (binsiz != 0) {
	    errmsg = "fBXNeʂI[o[܂";
	    goto xit;
	}

	/* D88t@Co */
	fp = fopen(d88fn, "wb");
	if (fp == NULL) {
	    errmsg = "D88t@C݂̏Ɏs܂";
	    goto xit;
	}
	fwrite(d88buf, 1, d88siz, fp);
	fclose(fp);
xit:
	if (d88buf != NULL) free(d88buf);
	if (errmsg != NULL) {
	    fprintf(stderr, "%s\n", errmsg);
	    return 0;
	}
//	fprintf(stdout, "Completed.\n");
	return 1;
}

int generate_pc88_d88(char *d88fn, unsigned char *bin, int binsiz, int loadaddr, int execaddr) {
	/* Write IPL */
	iplbin[2] =  loadaddr       & 0x00FF;
	iplbin[3] = (loadaddr >> 8) & 0x00FF;
	iplbin[4] =  binsiz         & 0x00FF;
	iplbin[5] = (binsiz   >> 8) & 0x00FF;
	iplbin[6] =  execaddr       & 0x00FF;
	iplbin[7] = (execaddr >> 8) & 0x00FF;
	if (!write_d88(d88fn, 0, 1, iplbin, sizeof(iplbin))) return 0;
	/* Write bin body */
	if (!write_d88(d88fn, 1, 1, bin, binsiz)) return 0;
	return 1;
}

unsigned char mz2000fat[256];
void mz2000fat_init(void) {
	int i;
	for (i=0; i<256; i++) mz2000fat[i] = 0;
	if (disk_type == DISK_TYPE_2DD) {
	    mz2000fat[1] = 0x30; /* f[^̈̊JnItZbg */
	    mz2000fat[2] = 0x01; /* gpς݃NX^ (FAT) */
	    mz2000fat[4] = 0x00; /* S̃NX^(500h) */
	    mz2000fat[5] = 0x05; /* S̃NX^(500h) */
	    mz2000fat[6] = 0x80; /* FAT̃NX^gpς݂ɂĂ */
	    mz2000fat[0xFF] = 1; /* 512bytes per cluster */
	    return;
	}
	mz2000fat[1] = 0x30; /* f[^̈̊JnItZbg */
	mz2000fat[2] = 0x01; /* gpς݃NX^ (FAT) */
	mz2000fat[4] = 0xA0; /* S̃NX^(4A0h) */
	mz2000fat[5] = 0x04; /* S̃NX^(4A0h) */
	mz2000fat[7] = 0x80; /* FAT̃NX^gpς݂ɂĂ */
}
void mz2000fat_setflag(int sector_number) {
	unsigned char d, m;
	if (disk_type == DISK_TYPE_2DD) {
	    d = mz2000fat[6 + (sector_number >> 4)];
	    m = 0x01 << ((sector_number >> 1) & 7);
            if (!(d & m)) {
		mz2000fat[6 + (sector_number >> 4)] |= m;
		mz2000fat[2]++;
		if (mz2000fat[2] == 0x00) mz2000fat[3]++;
	    }
	    return;
	}
	d = mz2000fat[6 + (sector_number >> 3)];
	m = 0x01 << (sector_number & 7);
	if (!(d & m)) {
	    mz2000fat[6 + (sector_number >> 3)] |= m;
	    mz2000fat[2]++;
	    if (mz2000fat[2] == 0x00) mz2000fat[3]++;
//	    mz2000fat[4]--;
//	    if (mz2000fat[4] == 0xFF) mz2000fat[5]--;
	}
}
int mz2000_sector_write(char *d88fn, unsigned char *bin, int sector_number) {
	int i, t, s;
	unsigned char tmp[256];
	for (i=0; i<256; i++) tmp[i] = ~bin[i];
	t = (sector_number >> 4) ^ 1;
	s = (sector_number & 0x000F) + 1;
	if (!write_d88(d88fn, t, s, tmp, 256)) return 0;
	mz2000fat_setflag(sector_number);
	return 1;
}

int generate_mz2000_d88(char *d88fn, unsigned char *bin, int binsiz, int loadaddr, int execaddr, char *bootname) {
	unsigned char tmp[256];
	int sector_number;
	int i;
	mz2000fat_init();
	/* Write IPL */
	for (i=0; i<256; i++) tmp[i] = 0;
	tmp[0] = 1; /* MZ-80B/2000/2200 */
	strcpy(&tmp[1], "IPLPRO");
	strcpy(&tmp[7], bootname);
	tmp[7 + strlen(bootname)] = 0x0D;
	tmp[0x14] =  binsiz         & 0x00FF;
	tmp[0x15] = (binsiz   >> 8) & 0x00FF;
	tmp[0x16] =  loadaddr       & 0x00FF;
	tmp[0x17] = (loadaddr >> 8) & 0x00FF;
	tmp[0x18] =  execaddr       & 0x00FF;
	tmp[0x19] = (execaddr >> 8) & 0x00FF;
	tmp[0x1D] = 0xFF; /* master disk */
	tmp[0x1E] = 0x30; /* boot sector */
	if (disk_type == DISK_TYPE_2DD) {
	    /* MZ-2500 2DD ubN08h`0Eh0000h`DFFFhɊ蓖Ă */
	    int siz, blk;
	    blk = ((loadaddr + 0x1FFF) >> 13) + 8;
	    for (i=0,siz=binsiz; siz>0; i++,siz-=0x2000) {
		tmp[0x20+i] = blk++;
	    }
	    tmp[0x20+i] = 0xFF;
	    for (i=0; i<7; i++) {
		tmp[0x30+i] = i + 8;
	    }
	    tmp[0x30+i] = 0xFF;
        }
	if (!mz2000_sector_write(d88fn, tmp, 0)) return 0;
	/* Write unused sector */
	for (i=0; i<256; i++) tmp[i] = 0xBF;
	for (i=0x01; i<=0x0E; i++) {
	    if (!mz2000_sector_write(d88fn, tmp, i)) return 0;
	}
	for (i=0x18; i<=0x2F; i++) {
	    if (!mz2000_sector_write(d88fn, tmp, i)) return 0;
	}
	/* Write directory */
	for (i=0; i<256; i++) tmp[i] = 0;
	for (i=0x10; i<=0x17; i++) {
	    if (!mz2000_sector_write(d88fn, tmp, i)) return 0;
	}
	/* Write binary body */
	sector_number = 0x30;
	for (; binsiz>=256; binsiz-=256) {
	    if (!mz2000_sector_write(d88fn, bin, sector_number++)) return 0;
	    bin += 256;
	}
	if (binsiz) {
	    for (i=0; i<binsiz; i++) tmp[i] = bin[i];
	    for (   ; i<256;    i++) tmp[i] = 0;
	    if (!mz2000_sector_write(d88fn, tmp, sector_number)) return 0;
	}
	/* flush FAT */
	if (!mz2000_sector_write(d88fn, mz2000fat, 0x0F)) return 0;
	return 1;
}

